%Nonlinear Discrete-Time Kalman Filter
%
% EKF implements an Extended Kalman Filter 
%  for estimation of frequency of encounter,
%  and wave spectrum modal frequency,
%  for a container ship, during parametric 
%  resonance, and plots the reults.
% 
% Box-Muller transformation 
%  is used for noise generation.
%
% A fourth order Runge-Kutta method is used 
%  for integration, with a step size of 0.1.
%
% See also: 
%  INIT, COMPUTE_F, COMPUTE_INPUT, RK4, BoxMuller

init;      % runs the init.m file
h = 0.1;   % sampling time
N = 10000; % no. of iterations
n = 8;     % number of states
m = 3;     % number of measurements
q = 0.1;   % std of process 
r = 0.1;   % std of measurement

% Design matrices
Q = q^0*eye(n);    % covariance of process
R = r^-1.7*eye(m); % covariance of measurement noise

% Initial conditions
x     = x_init;     % initial state 
P_bar = eye(n);     % initial state covariance 
x_hat = x;          % initial estimate
x_bar = zeros(n,1); % pre-allocating space
P_hat = zeros(n,n); % pre-allocating space

% Memory allocation
simdata = zeros(N+1,17);

%% MAIN LOOP %%
for i = 1:N+1,
 t = (i-1)*h; % time (sec)  
   
 % Modal wave frequency, slowly time-varying
%  x(4) = x(4) + 1e-7;

 % Measurement equation with noise
 a = [x(1) x(2) x(3)+x(8)]';
 z = BoxMuller(r,a);
 
%% Partial derivative Matrices
 % Jacobian
 A = [0,1,0,0,0,0,0,0;
    ((-const.C_44*(const.GM_m+const.GM_a*x_hat(5))/const.Ix_A44)+...
      (3*(x_hat(1)^2)*const.K_phi3/const.Ix_A44)) ...
      ,(const.B_44/const.Ix_A44),0,0 ...
      ,(-const.C_44*const.GM_a*x_hat(1)/const.Ix_A44),0,0,0;         
    0,0,-((const.B_11*(abs(x_hat(3))...
      +(x_hat(3)*sign(x_hat(3)))))/(const.m_A11))...
      ,0,0,0,0,0;
    0,0,0,0,0,0,0,0;
    0,0,0,0,0,1,0,0;
    0,0 ...
      ,(((-2*(x_hat(4)^3)*x_hat(5))/const.g)...
      -((x_hat(3)*(x_hat(4)^4)*x_hat(5))/(const.g^2))...
      -((x_hat(4)*x_hat(6)*const.B_11)*...
      ((x_hat(3)*x_hat(4)+const.g)*...
      (((abs(x_hat(3)))+x_hat(3)*sign(x_hat(3)))...
      -((abs(x_hat(3))))*x_hat(3)*x_hat(4))/...
      (((x_hat(3)*x_hat(4)+const.g)^2)*const.m_A11))))...
      ,(-(x_hat(4)*x_hat(5))-...
      ((2*x_hat(3)*(x_hat(4)^2)*x_hat(5))/const.g)-...
      (((x_hat(3)^2)*(x_hat(4)^3)*x_hat(5))/(const.g^2))-...
      ((x_hat(3)*x_hat(6)*const.B_11)/...
      ((x_hat(3)*x_hat(4)+const.g)*const.m_A11)))...
      ,(-(x_hat(4)^2)-((2*x_hat(3)*(x_hat(4)^3))/const.g)-...
      (((x_hat(3)^2)*(x_hat(4)^4))/(const.g^2)))...
      ,(-(x_hat(3)*x_hat(4)*const.B_11)/...
      ((x_hat(3)*x_hat(4)+const.g)*const.m_A11)),0,0;
    0,0,0,0,0,0,0,1;
    0,0,0,0,0,0,-const.omega_0^2,-2*const.lambda*const.omega_0];

 % Process noise matrix
 E = [1 0 0 0 0 0 0 0
      0 1 0 0 0 0 0 0
      0 0 1 0 0 0 0 0
      0 0 0 0 0 0 0 0
      0 0 0 0 1 0 0 0
      0 0 0 0 0 1 0 0
      0 0 0 0 0 0 1 0
      0 0 0 0 0 0 0 1]';

%% Discrete-time matrices: PHI, GAMMA, H
 [PHI,GAMMA] = c2d(A,E,h);
 H = [1 0 0 0 0 0 0 0             
      0 1 0 0 0 0 0 0          
      0 0 1 0 0 0 0 1];
     
%% PREDICT
 % Discrete-time extended KF model
 u   = compute_input(x,const);
 f_k = rk4(@(X_HAT,U) compute_f(X_HAT,U,const),x_hat,u,h);
 
 % Time update, predictor, (k+1)
 x_bar = f_k; % state estimate propagation
 P_bar = PHI*P_hat*PHI' + GAMMA*Q*GAMMA'; % error cov. prop.
 x     = rk4(@(x,u) compute_f(x,u,const),x,u,h); % integration

%% UPDATE
 % Estimation error (residual)
 z_bar = [x_bar(1) x_bar(2) x_bar(3)+x_bar(8)]';
 eps   = z - z_bar;
   
 % Measurement update, corrector
 K = ( P_bar*H' )/( H*P_bar*H' + R ); % EKF Kalman gain  
 x_hat = x_bar + K*eps; % state estimate update
 IKH   = eye(n) - K*H;
 P_hat = IKH*P_bar*IKH' + K*R*K'; % error cov. update
   
%% Store simulation data in a table
 simdata(i,:) = [t x' x_hat'];  
end

%% Simulation values
t         = simdata(:,1);
x_sim_EKF = simdata(:,2:9); 
x_est_EKF = simdata(:,10:17);

%% Plots
FontSize = 16.5; xaxis = t;
for j = 1:n,
 figure(j);
   set(gcf,'units','centimeters','pos',[0 0 20 12])
   set(gcf, 'Color', [1 1 1]); % sets figure background
   data1(j,:) = x_sim_EKF(:,j);
   data2(j,:) = x_est_EKF(:,j);
   data3(j,:) = data1(j,:)-data2(j,:);
 
 subplot(2,1,1);
   hLine1(j) = line(xaxis, data1(j,:));
   hLine2(j) = line(xaxis, data2(j,:));
   set(hLine1(j),'Color',[0.75 0 0]);
   set(hLine2(j),'Color',[0 0 0.80]);
   hLegend(j) = legend([hLine1(j), hLine2(j)], ...
      'Simulation','Estimate','location','Best');
  set(gca,'FontName','Helvetica');
  set([hLegend(j), gca],'FontSize',FontSize-6);
  hXLabel(j) = xlabel('Time [sec]');
  
 subplot(2,1,2);
   hLine3(j) = line(xaxis, data3(j,:));
   set(hLine3(j),'Color',[0.35 0.35 0.35]);
   set(gca,'FontName','Helvetica');
   set(gca,'TickDir','out','YGrid','on','LineWidth',1.0);
   hXLabel(j) = xlabel('Time [sec]');
end
 
%% FIGURE 1 %%
figure(1);
subplot(2,1,1); % plot state 1
hTitle  = title ('Roll angle'); 
hYLabel = ylabel('Roll amplitude [rad]');
set([hTitle, hXLabel, hYLabel],'FontName','Helvetica');
set([hXLabel, hYLabel],'FontSize',FontSize-4);
set(hTitle,'FontSize',FontSize,'FontWeight','bold');

subplot(2,1,2); % plot error state 1
hTitle  = title ('Roll angle error');
hYLabel = ylabel('Roll angle error [rad]');
set([hTitle, hXLabel, hYLabel],'FontName','Helvetica');
set([hXLabel, hYLabel],'FontSize',FontSize-4);
set(hTitle,'FontSize',FontSize,'FontWeight','bold');

%% FIGURE 2 %%
figure(2);
subplot(2,1,1); % plot state 2
hTitle  = title ('Roll rate');
hYLabel = ylabel('Rate of roll angle [rad/sec]');
set([hTitle, hXLabel, hYLabel],'FontName','Helvetica');
set([hXLabel, hYLabel],'FontSize',FontSize-4);
set(hTitle,'FontSize',FontSize,'FontWeight','bold');

subplot(2,1,2); % plot error state 2
hTitle  = title ('Roll rate error');
hYLabel = ylabel('Rate of roll angle error [rad/sec]');
set([hTitle, hXLabel, hYLabel],'FontName','Helvetica');
set([hXLabel, hYLabel],'FontSize',FontSize-4);
set(hTitle,'FontSize',FontSize,'FontWeight','bold');

%% FIGURE 3 %%
figure(3);
subplot(2,1,1); % plot state 3
hTitle  = title ('Ship forward speed'); 
hYLabel = ylabel('Speed [m/sec]');
set([hTitle, hXLabel, hYLabel],'FontName','Helvetica');
set([hXLabel, hYLabel],'FontSize',FontSize-4);
set(hTitle,'FontSize',FontSize,'FontWeight','bold');

subplot(2,1,2); % plot error state 3
hTitle  = title ('Ship forward speed error');
hYLabel = ylabel('Speed error [m/sec]');
set([hTitle, hXLabel, hYLabel],'FontName','Helvetica');
set([hXLabel, hYLabel],'FontSize',FontSize-4);
set(hTitle,'FontSize',FontSize,'FontWeight','bold');

%% FIGURE 4 %%
figure(4);
subplot(2,1,1); % plot state 4
hTitle  = title ('Modal wave frequency');
hYLabel = ylabel('Modal wave frequency [rad/sec]');
set([hTitle, hXLabel, hYLabel],'FontName','Helvetica');
set([hXLabel, hYLabel],'FontSize',FontSize-4);
set(hTitle,'FontSize',FontSize,'FontWeight','bold');

subplot(2,1,2); % plot error state 4
hTitle  = title ('Modal wave frequency error');
hYLabel = ylabel('Modal wave frequency error [rad/sec]');
set([hTitle, hXLabel, hYLabel],'FontName','Helvetica');
set([hXLabel, hYLabel],'FontSize',FontSize-4);
set(hTitle,'FontSize',FontSize,'FontWeight','bold');

%% FIGURE 5 %%
figure(5);
subplot(2,1,1); % plot state 5
hTitle  = title ('cos(integral[w_e(t)dt])');
hYLabel = ylabel('cos(integral[w_e(t)dt]) [rad/sec]');
set([hTitle, hXLabel, hYLabel],'FontName','Helvetica');
set([hXLabel, hYLabel],'FontSize',FontSize-4);
set(hTitle,'FontSize',FontSize,'FontWeight','bold');

subplot(2,1,2); % plot error state 5
hTitle  = title ('cos(integral[w_e(t)dt]) error');
hYLabel = ylabel('cos(integral[w_e(t)dt]) error [rad/sec]');
set([hTitle, hXLabel, hYLabel],'FontName','Helvetica');
set([hXLabel, hYLabel],'FontSize',FontSize-4);
set(hTitle,'FontSize',FontSize,'FontWeight','bold');

%% FIGURE 6 %%
figure(6);
subplot(2,1,1); % plot state 6
hTitle  = title ('Rate of cos(integral[w_e(t)dt])');
hYLabel = ylabel('d/dt{cos(integral[w_e(t)dt])} [rad/sec]');
set([hTitle, hXLabel, hYLabel],'FontName','Helvetica');
set([hXLabel, hYLabel],'FontSize',FontSize-4);
set(hTitle,'FontSize',FontSize,'FontWeight','bold');

subplot(2,1,2); % plot error state 6
hTitle  = title ('Rate of cos(integral[w_e(t)dt]) error');
hYLabel = ylabel('d/dt{cos(integral[w_e(t)dt])} error [rad/sec]');
set([hTitle, hXLabel, hYLabel],'FontName','Helvetica');
set([hXLabel, hYLabel],'FontSize',FontSize-4);
set(hTitle,'FontSize',FontSize,'FontWeight','bold');

%% FIGURE 7 %%
figure(7); 
subplot(2,1,1); % plot state 7
hTitle  = title ('Wave induced disturbance');
hYLabel = ylabel('Wave disturbance [rad/sec]');
set([hTitle, hXLabel, hYLabel],'FontName','Helvetica');
set([hXLabel, hYLabel],'FontSize',FontSize-4);
set(hTitle,'FontSize',FontSize,'FontWeight','bold');

subplot(2,1,2); % plot error state 7
hTitle  = title ('Wave induced disturbance error');
hYLabel = ylabel('Error [rad/sec]');
set([hTitle, hXLabel, hYLabel],'FontName','Helvetica');
set([hXLabel, hYLabel],'FontSize',FontSize-4);
set(hTitle,'FontSize',FontSize,'FontWeight','bold');

%% FIGURE 8 %%
figure(8);
subplot(2,1,1); % plot state 8
hTitle  = title ('Rate of wave induced disturbance');
hYLabel = ylabel('Rate of wave disturbance [rad/sec]');
set([hTitle, hXLabel, hYLabel],'FontName','Helvetica');
set([hXLabel, hYLabel],'FontSize',FontSize-4);
set(hTitle,'FontSize',FontSize,'FontWeight','bold');

subplot(2,1,2); % plot error state 8
hTitle  = title ('Rate of wave induced disturbance error');
hYLabel = ylabel('Error [rad/sec]');
set([hTitle, hXLabel, hYLabel],'FontName','Helvetica');
set([hXLabel, hYLabel],'FontSize',FontSize-4);
set(hTitle,'FontSize',FontSize,'FontWeight','bold');