%Nonlinear Kalman Filter
%
% UKF implements an Unscented Kalman Filter
%  for estimation of frequency of encounter,
%  and wave spectrum modal frequency,
%  for a container ship, during parametric 
%  resonance, and plots the reults.
%
% Box-Muller transformation 
%  is used for noise generation.
%
% A fourth order Runge-Kutta method is used 
%  for integration, with a step size of 0.1.
%
% See also: 
%  INIT, COMPUTE_F, COMPUTE_INPUT, 
%  RK4, BoxMuller, SIGMAS, UT_WEIGHTS

init;      % runs the init.m file
h = 0.1;   % sampling time
N = 10000; % no. of iterations
n = 8;     % number of states
m = 3;     % number of measurements
q = 0.1;   % std of process 
r = 0.1;   % std of measurement

% Design matrices
Q = q^3*eye(n); % covariance of process
R = r^1*eye(m); % covariance of measurement

% Initial conditions
x = x_init;         % initial state
x_hat_pluss_k = x;  % initial estimate
P_pluss_k = eye(n); % initial state covariance

% Unscented transformation weights
alpha = 1e-4; % transformation parameter                    
kappa = 0;    % transformation parameter                             
beta  = 2;    % transformation parameter 
[Wm,Wc,c] = ut_weights(n,alpha,kappa,beta); % weights
c = sqrt(c);

% Memory allocation
simdata = zeros(N+1,17);

%% MAIN LOOP %%
for i = 1:N+1,
 t = (i-1)*h; % time (sec)
  
 % Modal wave frequency, slowly time-varying
%  x(4) = x(4) + 1e-7;

 % Measurement equation with noise
 a = [x(1) x(2) x(3)+x(8)]';
 z = BoxMuller(r,a);
  
%% PREDICT
 % Sigma points around x_hat_pluss_k
 x_hat_k_1 = sigmas(x_hat_pluss_k,P_pluss_k,c);
  
 %Time update
 u = compute_input(x,const); % system input
 L = size(x_hat_k_1,2); 
 x_hat_k     = zeros(n,L); % pre-allocating space
 x_hat_bar_k = zeros(n,1); % pre-allocating space
  for k = 1:L
    % Sigma points propagated through transition funct. f.
    x_hat_k(:,k) = compute_f( x_hat_k_1(:,k),u,const );   
    % Combine vectors to obtain a priori state estimate
    x_hat_bar_k = x_hat_bar_k + ( Wm(k)*x_hat_k(:,k) );
  end
  
 % Estimate a priori error covariance
 P_bar_k = zeros(n,n);
  for k = 1:L
    x_hat_dev = x_hat_k(:,k)-x_hat_bar_k;
    P_bar_k = P_bar_k...
      + ( Wc(k) * (x_hat_dev * x_hat_dev') );
  end  
 % add Q, to take process noise into account
 P_bar_k = (P_bar_k + Q);   
  
 % Runge-Kutta 4th order integration
 x_hat_bar_k = ...
     rk4(@(X_Hat,U) compute_f(X_Hat,U,const),x_hat_pluss_k,u,h);
 x = rk4(@(X,U) compute_f(X,U,const),x,u,h);
  
%% UPDATE
 % Sigma points around x_hat_bar_k
 x_hat_k = sigmas(x_hat_bar_k,P_bar_k,c);

 % Measurement update
 y_hat_k     = zeros(m,L); % pre-allocating space
 y_hat_k_sum = zeros(m,1); % pre-allocating space
  for k = 1:L   
    % Sigma points projected through observation funct. h.
    y_hat_k(1,k) = x_hat_k(1,k);
    y_hat_k(2,k) = x_hat_k(2,k);
    y_hat_k(3,k) = x_hat_k(3,k) + x_hat_k(8,k);
    % Combine vectors to obtain predicted measurement
    y_hat_k_sum = y_hat_k_sum + (Wm(k) * y_hat_k(:,k));
  end
    
 % Covariance of predicted measurement
 P_y = zeros(m,m);
  for k = 1:L
    y_hat_dev = y_hat_k(:,k) - y_hat_k_sum;
    P_y = P_y + ( Wc(k)*( y_hat_dev*y_hat_dev' ) );
  end
 P_y = (P_y + R); % add R; take measurement noise into account
    
 % Cross-covariance of state and measurement
 P_xy = zeros(n,m);
  for k = 1:L
    x_hat_dev = x_hat_k(:,k)-x_hat_bar_k;
    y_hat_dev = y_hat_k(:,k) - y_hat_k_sum;
    P_xy = P_xy + ( Wc(k)*( x_hat_dev*y_hat_dev' ) );
  end
    
 % Measurement update
 K_k = P_xy / P_y; % UKF Kalman gain
 x_hat_pluss_k = x_hat_bar_k... % state estimate update
                + K_k * (z - y_hat_k_sum); 
 P_pluss_k = P_bar_k - K_k * P_y * K_k'; % cov. update
    
%% Store simulation data in a table
 simdata(i,:) = [t x' x_hat_pluss_k'];
end

%% Simulation values
t         = simdata(:,1); 
x_sim_UKF = simdata(:,2:9); 
x_est_UKF = simdata(:,10:17);

%% Plots 
FontSize = 16.5; xaxis = t;
for j = 1:n,
 figure(j);
   set(gcf,'units','centimeters','pos',[0 0 20 12])
   set(gcf, 'Color', [1 1 1]); % sets figure background
   data1(j,:) = x_sim_UKF(:,j);
   data2(j,:) = x_est_UKF(:,j);
   data3(j,:) = data1(j,:)-data2(j,:);
 
 subplot(2,1,1);
   hLine1(j) = line(xaxis, data1(j,:));
   hLine2(j) = line(xaxis, data2(j,:));
   set(hLine1(j),'Color',[0.75 0 0]);
   set(hLine2(j),'Color',[0 0 0.80]);
   hLegend(j) = legend([hLine1(j), hLine2(j)], ...
      'Simulation','Estimate','location','Best');
  set(gca,'FontName','Helvetica');
  set([hLegend(j), gca],'FontSize',FontSize-6);
  hXLabel(j) = xlabel('Time [sec]');
  
 subplot(2,1,2);
   hLine3(j) = line(xaxis, data3(j,:));
   set(hLine3(j),'Color',[0.35 0.35 0.35]);
   set(gca,'FontName','Helvetica');
   set(gca,'TickDir','out','YGrid','on','LineWidth',1.0);
   hXLabel(j) = xlabel('Time [sec]');
end

%% FIGURE 1 %%
figure(1);
subplot(2,1,1); % plot state 1
hTitle  = title ('Roll angle');
hYLabel = ylabel('Roll amplitude [rad]');
set([hTitle, hXLabel, hYLabel],'FontName','Helvetica');
set([hLegend, gca],'FontSize',FontSize-6);
set([hXLabel, hYLabel],'FontSize',FontSize-4);
set(hTitle,'FontSize',FontSize,'FontWeight','bold');

subplot(2,1,2); % plot error state 1
hTitle  = title ('Roll angle error');
hYLabel = ylabel('Roll angle error [rad]');
set([hTitle, hXLabel, hYLabel],'FontName','Helvetica');
set([hLegend, gca],'FontSize',FontSize-6);
set([hXLabel, hYLabel],'FontSize',FontSize-4);
set(hTitle,'FontSize',FontSize,'FontWeight','bold');

%% FIGURE 2 %%
figure(2);
subplot(2,1,1); % plot state 2
hTitle  = title ('Roll rate');
hYLabel = ylabel('Rate of roll angle [rad/sec]');
set([hTitle, hXLabel, hYLabel],'FontName','Helvetica');
set([hLegend, gca],'FontSize',FontSize-6);
set([hXLabel, hYLabel],'FontSize',FontSize-4);
set(hTitle,'FontSize',FontSize,'FontWeight','bold');

subplot(2,1,2); % plot error state 2
hTitle  = title ('Roll rate error');
hYLabel = ylabel('Rate of roll angle error [rad/sec]');
set([hTitle, hXLabel, hYLabel],'FontName','Helvetica');
set([hLegend, gca],'FontSize',FontSize-6);
set([hXLabel, hYLabel],'FontSize',FontSize-4);
set(hTitle,'FontSize',FontSize,'FontWeight','bold');

%% FIGURE 3 %%
figure(3);
subplot(2,1,1); % plot state 3
hTitle  = title ('Ship forward speed');
hYLabel = ylabel('Speed [m/sec]');
set([hTitle, hXLabel, hYLabel],'FontName','Helvetica');
set([hLegend, gca],'FontSize',FontSize-6);
set([hXLabel, hYLabel],'FontSize',FontSize-4);
set(hTitle,'FontSize',FontSize,'FontWeight','bold');

subplot(2,1,2); % plot error state 3
hTitle  = title ('Ship forward speed error');
hYLabel = ylabel('Speed error [m/sec]');
set([hTitle, hXLabel, hYLabel],'FontName','Helvetica');
set([hLegend, gca],'FontSize',FontSize-6);
set([hXLabel, hYLabel],'FontSize',FontSize-4);
set(hTitle,'FontSize',FontSize,'FontWeight','bold');

%% FIGURE 4 %%
figure(4);
subplot(2,1,1); % plot state 4
hTitle  = title ('Modal wave frequency');
hYLabel = ylabel('Modal wave frequency [rad/sec]');
set([hTitle, hXLabel, hYLabel],'FontName','Helvetica');
set([hLegend, gca],'FontSize',FontSize-6);
set([hXLabel, hYLabel],'FontSize',FontSize-4);
set(hTitle,'FontSize',FontSize,'FontWeight','bold');

subplot(2,1,2); % plot error state 4
hTitle  = title ('Modal wave frequency error');
hYLabel = ylabel('Modal wave frequency error [rad/sec]');
set([hTitle, hXLabel, hYLabel],'FontName','Helvetica');
set([hLegend, gca],'FontSize',FontSize-6);
set([hXLabel, hYLabel],'FontSize',FontSize-4);
set(hTitle,'FontSize',FontSize,'FontWeight','bold');

%% FIGURE 5 %%
figure(5);
subplot(2,1,1); % plot state 5
hTitle  = title ('cos(integral[w_e(t)dt])');
hYLabel = ylabel('cos(integral[w_e(t)dt]) [rad/sec]');
set([hTitle, hXLabel, hYLabel],'FontName','Helvetica');
set([hLegend, gca],'FontSize',FontSize-6);
set([hXLabel, hYLabel],'FontSize',FontSize-4);
set(hTitle,'FontSize',FontSize,'FontWeight','bold');

subplot(2,1,2); % plot error state 5
hTitle  = title ('cos(integral[w_e(t)dt]) error');
hYLabel = ylabel('cos(integral[w_e(t)dt]) error [rad/sec]');
set([hTitle, hXLabel, hYLabel],'FontName','Helvetica');
set([hLegend, gca],'FontSize',FontSize-6);
set([hXLabel, hYLabel],'FontSize',FontSize-4);
set(hTitle,'FontSize',FontSize,'FontWeight','bold');

%% FIGURE 6 %%
figure(6);
subplot(2,1,1); % plot state 6
hTitle  = title ('Rate of cos(integral[w_e(t)dt])');
hYLabel = ylabel('d/dt{cos(integral[w_e(t)dt])} [rad/sec]');
set([hTitle, hXLabel, hYLabel],'FontName','Helvetica');
set([hLegend, gca],'FontSize',FontSize-6);
set([hXLabel, hYLabel],'FontSize',FontSize-4);
set(hTitle,'FontSize',FontSize,'FontWeight','bold');

subplot(2,1,2); %Plot error state 6
hTitle  = title ('Rate of cos(integral[w_e(t)dt]) error');
hYLabel = ylabel('d/dt{cos(integral[w_e(t)dt])} error [rad/sec]');
set([hTitle, hXLabel, hYLabel],'FontName','Helvetica');
set([hLegend, gca],'FontSize',FontSize-6);
set([hXLabel, hYLabel],'FontSize',FontSize-4);
set(hTitle,'FontSize',FontSize,'FontWeight','bold');

 %% FIGURE 7 %%
figure(7); 
subplot(2,1,1); % plot state 7
hTitle  = title ('Wave induced disturbance');
hYLabel = ylabel('Wave disturbance [rad/sec]');
set([hTitle, hXLabel, hYLabel],'FontName','Helvetica');
set([hLegend, gca],'FontSize',FontSize-6);
set([hXLabel, hYLabel],'FontSize',FontSize-4);
set(hTitle,'FontSize',FontSize,'FontWeight','bold');

subplot(2,1,2); % plot error state 7
hTitle  = title ('Wave induced disturbance error');
hYLabel = ylabel('Error [rad/sec]');
set([hTitle, hXLabel, hYLabel],'FontName','Helvetica');
set([hLegend, gca],'FontSize',FontSize-6);
set([hXLabel, hYLabel],'FontSize',FontSize-4);
set(hTitle,'FontSize',FontSize,'FontWeight','bold');

%% FIGURE 8 %%
figure(8);
subplot(2,1,1); % plot state 8
hTitle  = title ('Rate of wave induced disturbance');
hYLabel = ylabel('Rate of wave disturbance [rad/sec]');
set([hTitle, hXLabel, hYLabel],'FontName','Helvetica');
set([hLegend, gca],'FontSize',FontSize-6);
set([hXLabel, hYLabel],'FontSize',FontSize-4);
set(hTitle,'FontSize',FontSize,'FontWeight','bold');

subplot(2,1,2); % plot error state 8
hTitle  = title ('Rate of wave induced disturbance error');
hYLabel = ylabel('Error [rad/sec]');
set([hTitle, hXLabel, hYLabel],'FontName','Helvetica');
set([hLegend, gca],'FontSize',FontSize-6);
set([hXLabel, hYLabel],'FontSize',FontSize-4);
set(hTitle,'FontSize',FontSize,'FontWeight','bold');