%Nonlinear Discrete-Time Kalman Filter
%
% Wave Direction Estimation
%  implements an extended Kalman filter
%  for estimation of the direction of the
%  incoming waves for a container ship
%  during parametric roll resonance.
%
% Box-Muller transformation 
%  is used for noise generation.
%
% A fourth order Runge-Kutta method is used 
%  for integration, with a step size of 0.1.
%
% See also: 
%   INIT, COMPUTE_INPUT, F_WAVE_DIR, RK4, 
%   BoxMuller, COMPUTE_RUDDER_ANGLE, NOMOTO

init;      % runs the init.m file
h = 0.1;   % sampling time
N = 10000; % no. of iterations
n = 9;     % number of states
m = 3;     % number of measurements
q = 0.1;   % std of process 
r = 0.1;   % std of measurement

% Design matrices
Q = q^0*eye(n);    % covariance of process
R = r^-1.7*eye(m); % covariance of measurement noise

% Initial conditions
x = x_init_wave_dir; % initial state 
P_bar = eye(n);      % initial state covariance 
x_hat = x;           % initial estimate
u     = [0 0 5]';    % initial input vector
x_bar = zeros(n,1);  % pre-allocating space
P_hat = zeros(n,n);  % pre-allocating space

% Memory allocation
simdata = zeros(N+1,19);

%% MAIN LOOP
for i = 1:N+1,
 t = (i-1)*h; % time (sec)  
 
 % Wave direction (beta), slowly time-varying
%  x(9) = x(9) + 1e-5;

 % Measurement equation with noise
 a = [x(1) x(2) x(3)+x(8)]';
 z = BoxMuller(r,a);
 
 % Rudder angle
 delta = compute_rudder_angle(u,const);
 
 % System input vector
 tau_1 = compute_input(x,const);
 u_temp = rk4(@(u,delta) nomoto(u,delta,const),u,delta,h);
 psi = BoxMuller(r,u_temp(2,1)); 
%   if i>1000, 
%    psi_c = 20*D2R; % 20 deg heading command
%   else
%    psi_c = 0;
%   end
%  psi = BoxMuller(r,psi_c);
 u = [tau_1 psi u_temp(3,1)]';
   
%% Partial derivative Matrices
 % Variables for use in the Jacobian Matrix
 cos1 = cos(x_hat(1)); 
 sin1 = sin(x_hat(1));
 cos_x9u2 = cos(x_hat(9)-u(2)); 
 sin_x9u2 = sin(x_hat(9)-u(2));
 var = (u(1)-const.B_11*x_hat(3)*abs(x_hat(3)))/const.m_A11;
 var2 = (x_hat(4)+((x_hat(4)^2)*x_hat(3)*cos_x9u2))/const.g;
 var3 = -const.B_11*(abs(x_hat(3)))/const.m_A11...
        -(const.B_11*x_hat(3)*sign(x_hat(3)))/const.m_A11;
 var4 = const.g-x_hat(3)*x_hat(4)*cos_x9u2;
 
 % Jacobian
 A = [0,1,0,0,0,0,0,0,0;
    ((-const.C_44*(const.GM_m+const.GM_a*x_hat(5))/const.Ix_A44)...
      +(3*(x_hat(1)^2)*const.K_phi3/const.Ix_A44))...
      ,(const.B_44/const.Ix_A44),0,0 ...
      ,(-const.C_44*const.GM_a*x_hat(1)/const.Ix_A44),0,0,0,0;         
    0,0,var3,0,0,0,0,0,0;
    0,0,0,0,0,0,0,0,0;
    0,0,0,0,0,1,0,0,0;
    -((sin1*x_hat(3)*x_hat(4)*x_hat(6)*u(3)*sin_x9u2)/var4)...
      ,0 ...
      ,(-(2*x_hat(5)*var2*(x_hat(4)^2)*cos_x9u2/const.g)...
       -((x_hat(4)*x_hat(6)*var3*cos_x9u2)/var4)...
       -(((x_hat(4)^2)*x_hat(6)*var*(cos_x9u2^2))/(var4^2))... 
       +((cos1*x_hat(4)*x_hat(6)*u(3)*sin_x9u2)/var4)...
       +((cos1*x_hat(3)*(x_hat(4)^2)*x_hat(6)*u(3)...
          *sin_x9u2*cos_x9u2)/(var4^2)))...
      ,(-2*x_hat(5)*var2*(1+(2*x_hat(3)*x_hat(4)*cos_x9u2/const.g))...
       -((x_hat(6)*var*cos_x9u2)/var4)...
       -((x_hat(4)*x_hat(6)*var*(cos_x9u2^2)*x_hat(3))/(var4^2))...
       +((cos1*x_hat(3)*x_hat(6)*u(3)*sin_x9u2)/var4)...
       +((cos1*(x_hat(3)^2)*x_hat(4)*x_hat(6)*u(3)...
          *sin_x9u2*cos_x9u2)/(var4^2)))...
      ,(-(var2^2))...
      ,(-((x_hat(4)*var*cos_x9u2)/var4)...
       +((cos1*x_hat(3)*x_hat(4)*u(3)*sin_x9u2)/var4))...
      ,0,0 ...
      ,(-((2*x_hat(5)*var2*(x_hat(4)^2)*x_hat(3)*sin_x9u2)/const.g)...
       -((x_hat(4)*x_hat(6)*var*sin_x9u2)/var4)...
       -(((x_hat(4)^2)*x_hat(6)*var*cos_x9u2...
          *x_hat(3)*sin_x9u2)/(var4^2))...
       -((cos1*x_hat(3)*x_hat(4)*x_hat(6)*u(3)*cos_x9u2)/var4)...
       +((cos1*(x_hat(3)^2)*(x_hat(4)^2)*x_hat(6)...
          *u(3)*(sin_x9u2^2))/(var4^2)));
      0,0,0,0,0,0,0,1,0;
      0,0,0,0,0,0,-const.omega_0^2,-2*const.lambda*const.omega_0,0;
      0,0,0,0,0,0,0,0,0];

 % Process noise matrix
 E = [1 0 0 0 0 0 0 0 0
      0 1 0 0 0 0 0 0 0
      0 0 1 0 0 0 0 0 0
      0 0 0 0 0 0 0 0 0
      0 0 0 0 1 0 0 0 0
      0 0 0 0 0 1 0 0 0
      0 0 0 0 0 0 1 0 0
      0 0 0 0 0 0 0 1 0
      0 0 0 0 0 0 0 0 1]';

%% Discrete-time matrices: PHI, GAMMA, H
 [PHI,GAMMA] = c2d(A,E,h);
 H = [1 0 0 0 0 0 0 0 0            
      0 1 0 0 0 0 0 0 0          
      0 0 1 0 0 0 0 1 0];
     
%% PREDICT
 % Time update, predictor
 x_bar = ... % state estimate propagation
         rk4(@(X_HAT,U) f_wave_dir(X_HAT,U,const),x_hat,u,h); 
 P_bar = PHI*P_hat*PHI' + GAMMA*Q*GAMMA'; % error covariance prop.
 x     = rk4(@(x,u) f_wave_dir(x,u,const),x,u,h); % integration
 
%% UPDATE
 % Estimation error (residual)
 z_bar = [x_bar(1) x_bar(2) x_bar(3)+x_bar(8)]';
 eps   = z - z_bar;
 
 % Measurement update, corrector
 K     = ( P_bar*H' )/( H*P_bar*H' + R ); % EKF Kalman gain  
 x_hat = x_bar + K*eps; % state estimate update
 IKH   = eye(n) - K*H;
 P_hat = IKH*P_bar*IKH' + K*R*K'; % error covariance update
   
%% Store simulation data in a table
 simdata(i,:) = [t x' x_hat'];
end

%% Simulation values
t         = simdata(:,1);
x_sim_EKF = simdata(:,2:10); 
x_est_EKF = simdata(:,11:19);

%% Plot of wave direction estimation %%
FontSize = 16.5; FontName = 'Helvetica';

% Data you want to plot
data1 = x_sim_EKF(:,9); xaxis = t; 
data2 = x_est_EKF(:,9); data3 = (data1-data2);

figure;
  set(gcf,'units','centimeters','pos', [0 0 20 12])
  set(gcf, 'Color', [1 1 1]); % Sets figure background
  set(gca, 'Color', [1 1 1]); % Sets axes background
  set(gcf, 'Renderer', 'painters'); 

subplot(2,1,1)
hLine1 = line(xaxis, data1); 
hLine2 = line(xaxis, data2);
set(hLine1,'LineWidth',2,'Color',[0.75 0 0]);
set(hLine2,'LineWidth',2,'Color',[0 0 0.80]);

% Title, and names on the axes
hTitle  = title ('Direction of incoming waves');
hXLabel = xlabel('Time [sec]');
hYLabel = ylabel('Direction of waves [rad]');

hLegend = legend([hLine1, hLine2], ...
      'Simulation','Estimate','location','Best');

set(gca,'FontName',FontName);
set([hTitle, hXLabel, hYLabel],'FontName',FontName);
set([hLegend, gca],'FontSize',FontSize-6);
set([hXLabel, hYLabel],'FontSize',FontSize-4);
set(hTitle,'FontSize',FontSize,'FontWeight','bold');

set(gca,'TickDir','out','TickLength',[.02 .02],...
    'LineWidth',1.0);

subplot(2,1,2)  
hLine3 = line(xaxis, data3);
set(hLine3,'LineWidth',2,'Color',[0.35 0.35 0.35]);

hTitle  = title ('Direction of incoming waves error');
hXLabel = xlabel('Time [sec]');
hYLabel = ylabel('Error [rad]');
set(gca,'FontName',FontName);
set([hTitle, hXLabel, hYLabel],'FontName',FontName);
set([hLegend, gca],'FontSize',FontSize-6);
set([hXLabel, hYLabel],'FontSize',FontSize-4);
set(hTitle,'FontSize',FontSize,'FontWeight','bold');

set(gca,'TickDir','out','TickLength',[.02 .02], ...
    'YGrid','on','XColor',[.0 .0 .0], ...
    'YColor',[.0 .0 .0],'LineWidth',1.0);